import os
import shutil
import black
import logging
import yaml

from jinja2 import Template

from pystorz.store import utils
from pystorz.mgen.utils import yaml_files
from pystorz.mgen.meta import Struct, Resource


log = logging.getLogger(__name__)
log.debug("loading generator.py...")


def Generate(*models) -> None:
    structs = dict()
    resources = dict()

    for model in models:
        log.debug(f"Loading model {model}...")

        s, r = _load_model(model)

        structs.update(s)
        resources.update(r)

    res = _cleanup(
        _render(
            "mgen/templates/python.py",
            resources,
            _dependency_order(structs)))    

    # refactor and format python code
    res = _reformat_python_code(res)

    targetDir = "generated"

    if os.path.exists(targetDir):
        # delete the directory with contents
        shutil.rmtree(targetDir)

    utils.export_file(targetDir, "model.py", res)


    res = _render(
            "mgen/templates/javascript.js",
            resources,
            _dependency_order(structs))

    # refactor and format javascript code
    targetDir = "generated"

    utils.export_file(targetDir, "model.js", res)

    

def _load_model(path: str):
    yamls = yaml_files(path)
    structs = dict()
    resources = dict()

    for y in yamls:
        model = _read_model(y)

        for m in model["types"]:
            if m["kind"].lower() == "struct":
                if m["name"] in structs:
                    raise Exception("duplicate struct: {}".format(m["name"]))
                structs[m["name"]] = Struct(m)
                continue

            if m["kind"].lower() == "object":
                if m["name"] in resources:
                    raise Exception("duplicate object: {}".format(m["name"]))
                resources[m["name"]] = Resource(m)
                continue

            raise Exception("unknown kind: {}".format(m["kind"]))

    errors = _validate_model(structs, resources)
    if len(errors) > 0:
        raise Exception(
            """model validation failed:
    {}""".format(
                """
    """.join(
                    errors
                )
            )
        )

    return structs, resources


def _validate_model(structs, resources):
    errors = []
    known_types = ["string", "int", "float", "bool", "datetime"]

    for _, s in structs.items():
        known_types.append(s.name)

    for _, s in structs.items():
        for p in s.properties:
            if p.IsArray():
                elem_type = p.SubType()
                if elem_type not in known_types:
                    errors.append(
                        "struct {} property {}: unknown type: {}".format(
                            s.name, p.name, elem_type
                        )
                    )
                continue

            if p.IsMap():
                # map[int]string
                elem_type = p.SubType()
                key_type = p.type[4:].split("]")[0]
                val_type = p.type[4:].split("]")[1]
                if key_type not in known_types:
                    errors.append(
                        "struct {} property {}: unknown type: {}".format(
                            s.name, p.name, elem_type
                        )
                    )
                if val_type not in known_types:
                    errors.append(
                        "struct {} property {}: unknown type: {}".format(
                            s.name, p.name, elem_type
                        )
                    )
                continue

            if p.type not in known_types:
                errors.append(
                    "struct {} property {}: unknown type: {}".format(
                        s.name, p.name, p.type
                    )
                )

    for _, r in resources.items():
        if r.external is None and r.internal is None:
            errors.append("resource {} has no internal and external".format(r.name))
            continue

        if r.external is not None:
            if r.external not in known_types:
                errors.append(
                    "resource {} external: unknown type: {}".format(r.name, r.external)
                )

        if r.internal is not None:
            if r.internal not in known_types:
                errors.append(
                    "resource {} internal: unknown type: {}".format(r.name, r.internal)
                )

    return errors


def _read_model(path: str):
    log.debug(f"reading model from {path}")

    with open(path, "r") as f:
        data = yaml.safe_load(f)
    return data


def _dependency_order(structs) -> list[Struct]:
    all_structs = {}
    dependencies = {}
    for _, s in structs.items():
        all_structs[s.name] = s
        dependencies[s.name] = set()

    known_types = ["string", "int", "float", "bool", "datetime"]

    for _, s in structs.items():
        for p in s.properties:
            if p.IsArray():
                elem_type = p.SubType()
                if elem_type not in known_types:
                    dependencies[s.name].add(elem_type)

                continue

            if p.IsMap():
                # map[string]bla
                elem_type = p.SubType()
                key_type = p.type[4:].split("]")[0]
                if key_type not in known_types:
                    raise Exception(
                        f"Invalid type {key_type} in {p.type}. Please use one of {known_types} as map key type"
                    )

                val_type = p.type[4:].split("]")[1]
                if val_type not in known_types:
                    dependencies[s.name].add(val_type)

                continue

            if p.type not in known_types:
                dependencies[s.name].add(p.type)

    ordered = []
    while len(ordered) < len(structs):
        did_stuff = False
        for _, s in structs.items():
            if s in ordered:
                continue

            if len(dependencies[s.name]) == 0:
                ordered.append(s)
                for k in dependencies.keys():
                    if s.name in dependencies[k]:
                        dependencies[k].remove(s.name)
                        did_stuff = True

        if not did_stuff and len(ordered) < len(structs):
            log.debug(f"dependencies: {dependencies}")
            log.debug(f"ordered: {[ o.name for o in ordered]}")
            log.debug(f"lengths: {len(dependencies)} {len(ordered)} {len(structs)}")

            raise Exception("Circular dependency detected")

    return ordered


def _render(rpath: str, resources, structs) -> str:
    path = os.path.join(utils.runtime_dir(), rpath)

    with open(path, "r") as f:
        content = f.read()

    t = Template(content)
    return t.render(resources=resources, structs=structs)


def _reformat_python_code(code_str):
    try:
        # Reformat the code string using the parsed AST.
        formatted_code = black.format_str(code_str, mode=black.FileMode())

        return formatted_code

    except Exception as e:
        log.error("failed to reformat code:", e)
        return code_str

def _cleanup(code_str):
    # remove empty lines
    code_str = code_str.replace("&#34;", '"')
    return os.linesep.join([s for s in code_str.splitlines() if s.strip()])



import logging
from pymongo import MongoClient, ASCENDING, DESCENDING

from pystorz.internal import constants
from pystorz.store import store, options, utils


log = logging.getLogger(__name__)

COLLECTION_NAME = "objects"
DATABASE_NAME = "pystorz"
TIMEOUT = 10  # seconds


def _convert_filter(filterOption, sample):
    if isinstance(filterOption, options._ListDeleteOption):
        copt = options.CommonOptionHolderFactory()
        filterOption.ApplyFunction()(copt)
        filterOption = copt.filter

    if isinstance(filterOption, options.AndOption):
        return {"$and": [_convert_filter(f, sample) for f in filterOption.filters]}

    if isinstance(filterOption, options.OrOption):
        return {"$or": [_convert_filter(f, sample) for f in filterOption.filters]}

    if isinstance(filterOption, options.NotOption):
        return {"$nor": [_convert_filter(filterOption.filter, sample)]}

    if utils.object_path(sample, filterOption.key) is None:
        raise Exception(constants.ErrInvalidFilter)

    def convert_value(v):
        # return "'{}'".format(v)
        if isinstance(v, str):
            return utils.encode_string(v)
        return v

    if isinstance(filterOption, options.InOption):
        return {
            f"object.{filterOption.key}": {
                "$in": [convert_value(v) for v in filterOption.values]
            }
        }

    if isinstance(filterOption, options.EqOption):
        return {f"object.{filterOption.key}": filterOption.value}

    if isinstance(filterOption, options.LtOption):
        return {f"object.{filterOption.key}": {"$lt": filterOption.value}}

    if isinstance(filterOption, options.GtOption):
        return {f"object.{filterOption.key}": {"$gt": filterOption.value}}

    if isinstance(filterOption, options.LteOption):
        return {f"object.{filterOption.key}": {"$lte": filterOption.value}}

    if isinstance(filterOption, options.GteOption):
        return {f"object.{filterOption.key}": {"$gte": filterOption.value}}

    raise Exception(constants.ErrInvalidFilter)


class MongoStore(store.Store):
    def __init__(self, Schema, URI: str):
        self._schema = Schema
        self._URI = URI
        self._client = None
        self._db = DATABASE_NAME
        self._test_connection()

    def _test_connection(self):
        if self._client is not None:
            try:
                self._client.admin.command("ping")
                return
            except Exception:
                pass
        self._client = MongoClient(self._URI, serverSelectionTimeoutMS=TIMEOUT * 1000)
        self._client.admin.command("ping")
        self._prepare()

    def _prepare(self):
        collection = self._client[self._db][COLLECTION_NAME]
        collection.create_index([("idpath", ASCENDING)])
        collection.create_index([("pkpath", ASCENDING)])
        collection.create_index([("type", ASCENDING)])

    def Create(self, obj: store.Object, *opt: options.CreateOption) -> store.Object:
        if obj is None:
            raise Exception(constants.ErrObjectNil)
        log.info(f"create {obj.Metadata().Kind()} {obj.PrimaryKey()}")
        lk = obj.Metadata().Kind().lower()
        path = f"{lk}/{obj.PrimaryKey()}"
        existing = None
        try:
            existing = self.Get(store.ObjectIdentity(path))
        except Exception:
            pass
        if existing is not None:
            raise Exception(constants.ErrObjectExists)
        self._test_connection()
        typ = obj.Metadata().Kind().lower()
        collection = self._client[self._db][COLLECTION_NAME]
        record = {
            "idpath": obj.Metadata().Identity().Path(),
            "pkpath": f"{typ}/{obj.PrimaryKey()}",
            "pkey": obj.PrimaryKey(),
            "type": typ,
            "object": obj.ToDict(),
        }
        collection.insert_one(record)
        return obj.Clone()

    def Update(
        self,
        identity: store.ObjectIdentity,
        obj: store.Object,
        *opt: options.UpdateOption,
    ) -> store.Object:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        if obj is None:
            raise Exception(constants.ErrObjectNil)
        log.info(f"update {identity.Path()}")
        existing = self.Get(identity)
        if existing.Metadata().Kind() != obj.Metadata().Kind():
            raise Exception(constants.ErrObjectIdentityMismatch)
        if existing.PrimaryKey() != obj.PrimaryKey():
            if (
                existing.Metadata().Identity().Path()
                != obj.Metadata().Identity().Path()
            ):
                raise Exception(constants.ErrObjectIdentityMismatch)
            target_identity = store.ObjectIdentity(
                f"{existing.Metadata().Kind().lower()}/{obj.PrimaryKey()}"
            )
            target = None
            try:
                target = self.Get(target_identity)
            except Exception:
                pass
            if target is not None:
                raise Exception(constants.ErrObjectExists)
        self._test_connection()
        self.Delete(identity)
        return self.Create(obj)

    def Delete(self, identity: store.ObjectIdentity, *opt: options.DeleteOption):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info(f"delete {identity.Path()}")
        copt = options.CommonOptionHolderFactory()
        for o in opt:
            o.ApplyFunction()(copt)
        self._test_connection()
        collection = self._client[self._db][COLLECTION_NAME]
        if copt.filter is None:
            if self.Get(identity) is None:
                raise Exception(constants.ErrNoSuchObject)
            if identity.IsId():
                collection.delete_many({"idpath": identity.Path()})
            else:
                collection.delete_many({"pkpath": identity.Path()})
        else:
            obj = self._schema.ObjectForKind(identity.Type())
            if obj is None:
                raise Exception(constants.ErrNoSuchObject)

            filter_ = {
                "$and": [{"type": identity.Type()},
                         _convert_filter(copt.filter, obj)]
            }
            
            log.info(f"filter: {filter_}")
            collection.delete_many(filter_)

    def Get(
        self, identity: store.ObjectIdentity, *opt: options.GetOption
    ) -> store.Object:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info(f"get {identity.Path()}")
        copt = options.CommonOptionHolderFactory()
        for o in opt:
            o.ApplyFunction()(copt)
        self._test_connection()
        collection = self._client[self._db][COLLECTION_NAME]
        res = None
        key = "idpath" if identity.IsId() else "pkpath"
        res = collection.find_one({key: identity.Path()})

        if res is None:
            raise Exception(constants.ErrNoSuchObject)

        resource = self._schema.ObjectForKind(res["type"])
        resource.FromDict(res["object"])
        return resource

    def List(
        self, identity: store.ObjectIdentity, *opt: options.ListOption
    ) -> store.ObjectList:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info(f"list {identity.Path()}")
        if len(identity.Key()) > 0:
            raise Exception(constants.ErrInvalidPath)
        copt = options.CommonOptionHolderFactory()
        for o in opt:
            o.ApplyFunction()(copt)

        self._test_connection()
        collection = self._client[self._db][COLLECTION_NAME]
        filter_ = {"type": identity.Type()}

        if copt.filter:
            obj = self._schema.ObjectForKind(identity.Type())
            if obj is None:
                raise Exception(constants.ErrNoSuchObject)

            filter_ = {"$and": [filter_, _convert_filter(copt.filter, obj)]}
            log.info(f"filter: {filter_}")

        cur = collection.find(filter_)

        if copt.order_by:
            cur = cur.sort(
                f"object.{copt.order_by}",
                ASCENDING if copt.order_incremental else DESCENDING,
            )

        if copt.page_size is not None and copt.page_size > 0:
            cur = cur.limit(int(copt.page_size))
        if copt.page_offset is not None and copt.page_offset > 0:
            cur = cur.skip(int(copt.page_offset))

        rows = list(cur)
        res = store.ObjectList()
        for r in rows:
            try:
                resource = self._schema.ObjectForKind(identity.Type())
                resource.FromDict(r["object"])

                res.append(resource)
            except Exception as e:
                log.error(str(e))
                continue
        return res


def MongoStoreFactory(schema: store.SchemaHolder, URI: str):
    return MongoStore(schema, URI)
    


import json
from pystorz.internal import constants
# "from pystorz.store import utils",
from pystorz.store import store
from datetime import datetime
from typing import Type


{% for data in structs %}

class {{ data.name }}:

	def __init__(self):
		raise Exception("cannot initialize like this. use the factory method")
	
	def ToDict(self):
		raise Exception("not implemented")

	def FromDict(self, data):
		raise Exception("not implemented")

{% for prop in data.properties %}
	def {{ prop.CapitalizedName() }}(self) -> {{ prop.StrippedType() }}:
		raise Exception("not implemented")

	def Set{{ prop.CapitalizedName() }}(self, val: {{prop.StrippedType()}}):
		raise Exception("not implemented")
{% endfor %}


def {{ data.name }}Factory() -> {{ data.name }}:
	ret = _{{ data.name }}()
	{% for prop in data.properties %}
	ret.{{ prop.name }}_ = {{ prop.Default() }}
	{% endfor %}
	return ret


class _{{ data.name }}({{data.name}}):

	def __init__(self):
		{% for prop in data.properties %}
		self.{{ prop.name }}_ = {{prop.Default()}}
		{% endfor %}

	{% for prop in data.properties %}
	def Set{{ prop.CapitalizedName() }}(self, val):
		{% if prop.type == "datetime" %}
		self.{{ prop.name }}_ = store.datetime_string(val)
		{% elif prop.type == "string" %}
		self.{{ prop.name }}_ = str(val)
		{% elif prop.type == "int" %}
		self.{{ prop.name }}_ = int(val)
		{% elif prop.type == "float" %}
		self.{{ prop.name }}_ = float(val)
		{% elif prop.type == "bool" %}
		self.{{ prop.name }}_ = bool(val)
		{% else %}
		self.{{ prop.name }}_ = val
		{% endif %}

	def {{ prop.CapitalizedName() }}(self):
		{% if prop.type == "datetime" %}
		return store.datetime_parse(self.{{ prop.name }}_)
		{% else %}
		return self.{{ prop.name }}_
		{% endif %}

	{% endfor %}

	
	def FromJson(self, jstr):
		data = json.loads(jstr)
		return self.FromDict(data)


	def ToJson(self):
		return json.dumps(self.ToDict())


	def ToDict(self):
		data = {}
		{% for prop in data.properties %}
		{% if prop.IsArray() %}
		rawList = []
		for v in self.{{prop.name}}_:
			{% if prop.IsComplexType() %}
			rawList.append(v.ToDict())
			{% else %}
			rawList.append(v)
			{% endif %}

		data["{{prop.name}}"] = rawList
		{% elif prop.IsMap() %}
		rawSubmap = {}
		for k, v in self.{{prop.name}}_.items():
			{% if prop.IsComplexType() %}
			rawSubmap[k] = v.ToDict()
			{% else %}
			rawSubmap[k] = v
			{% endif %}

		data["{{prop.name}}"] = rawSubmap
		{% else %}
		
		{% if prop.IsComplexType() %}
		# if self.{{prop.name}}_ is not None:
		data["{{prop.name}}"] = self.{{prop.name}}_.ToDict()
		{% else %}
		data["{{prop.name}}"] = self.{{prop.name}}_
		{% endif %}
	
		{% endif %}
		{% endfor %}
		return data
	

	def FromDict(self, data):
		for key, rawValue in data.items():
			if rawValue is None:
				continue
			
			{% for prop in data.properties %}
			if key == "{{prop.name}}":
				{% if prop.IsArray() %}
				res = {{ prop.Default() }}

				for rw in rawValue:
					ud = {{ prop.ComplexTypeValueDefault() }}
					{% if prop.IsComplexType() %}
					ud.FromDict(rw)
					{% else %}
					ud = rw
					{% endif %}
					res.append(ud)

				self.{{prop.name}}_ = res
				{% elif prop.IsMap() %}
				res = {{ prop.Default() }}
				
				for rk, rw in rawValue.items():
					ud = {{prop.ComplexTypeValueDefault()}}
					{% if prop.IsComplexType() %}
					ud.FromDict(rw)
					{% else %}
					ud = rw
					{% endif %}
					res[rk] = ud

				self.{{prop.name}}_ = res
				{% else %}
				{% if prop.IsComplexType() %}
				self.{{prop.name}}_.FromDict(rawValue)
				{% else %}
				self.{{prop.name}}_ = rawValue
				{% endif %}
				
				{% endif %}
			{% endfor %}
		
{% endfor %}


{% for _, data in resources.items() %}

{% if data.external %}
class {{ data.name }}(store.ExternalHolder):
{% else %}
class {{ data.name }}(store.Object):
{% endif %}

	def __init__(self):
		raise Exception("cannot initialize like this. use the factory method")
	
	def ToDict(self):
		raise Exception("not implemented")

	def FromDict(self, data):
		raise Exception("not implemented")

	def Clone(self) -> Type["{{data.name}}"]:
		raise NotImplementedError()

	def Meta(self) -> store.Meta:
		raise Exception("not implemented")

	{% if data.external %}
	def External() -> {{ data.external }}:
		raise Exception("not implemented")

	{% endif %}

	{% if data.internal %}
	def Internal(self) -> {{ data.internal }}:
		raise Exception("not implemented")

	{% endif %}



def {{ data.name }}Factory() -> {{ data.name }}:
	ret = _{{ data.name }}()

	{% if data.external %}
	ret.external_ = {{ data.external }}Factory()
	{% endif %}
	{% if data.internal %}
	ret.internal_ = {{ data.internal }}Factory()
	{% endif %}

	return ret


class _{{ data.name }}({{data.name}}):

	def __init__(self):
		self.meta_ = store.MetaFactory("{{ data.name }}")
		self.external_ = None
		self.internal_ = None

	{% if data.external %}
	def SetExternal(self, val):
		self.external_ = val

	def External(self) -> {{ data.external }}:
		return self.external_

	{% endif %}

	{% if data.internal %}
	def SetInternal(self, val):
		self.internal_ = val
	
	def Internal(self) -> {{ data.internal }}:
		return self.internal_

	{% endif %}

	
	def FromJson(self, jstr):
		data = json.loads(jstr)
		return self.FromDict(data)


	def ToJson(self):
		return json.dumps(self.ToDict())


	def ToDict(self):
		data = {}
		data["metadata"] = self.meta_.ToDict()
		{% if data.external %}data["external"] = self.external_.ToDict(){% endif %}
		{% if data.internal %}data["internal"] = self.internal_.ToDict(){% endif %}
		return data


	def FromDict(self, data):
		for key, rawValue in data.items():
			if rawValue is None:
				continue
			
			if key == "metadata":
				self.meta_.FromDict(rawValue)

			{% if data.external %}
			if key == "external":
				self.external_.FromDict(rawValue)
			{% endif %}

			{% if data.internal %}
			if key == "internal":
				self.internal_.FromDict(rawValue)
			{% endif %}
	
	def Clone(self) -> {{data.name}}:
		ret = {{data.name}}Factory()
		ret.FromJson(self.ToJson())
		return ret

	def Metadata(self) -> store.Meta:
		return self.meta_

	def SetMetadata(self, val: store.Meta):
		self.meta_ = val

	def PrimaryKey(self):
		return str(self.{{data.PrimaryKeyFunctionCaller()}})


def {{data.name}}Identity(pkey):
	return store.ObjectIdentity("{{ data.IdentityPrefix() }}/" + pkey )

{{data.name}}KindIdentity = store.ObjectIdentity("{{ data.IdentityPrefix() }}/")

{{data.name}}Kind = "{{ data.name }}"


{% endfor %}


 

class _Schema(store.SchemaHolder):
	def __init__(self, objects):
		self.objects = objects
	
	def ObjectForKind(self, kind) -> store.Object:
		{% for _, r in resources.items() %}
		
		if kind == "{{r.name}}":
			return {{r.name}}Factory()
		elif kind == "{{r.IdentityPrefix()}}":
			return {{r.name}}Factory()

		{% endfor %}
		
		raise Exception(constants.ErrNoSuchObject)
	
	def Types(self):
		return self.objects


def Schema():
	objects = [
		{% for _, r in resources.items() %}
		"{{r.name}}",
		{% endfor %}
	]

	return _Schema(objects)

 


from pystorz.mgen.utils import capitalize


BASIC_TYPES = ["string", "int", "float", "bool", "datetime"]


def typeDefault(tp: str) -> str:
    if tp.startswith("list") or tp.startswith("[]"):
        return "list()"
        # return f"{tp} {{}}"
    if tp.startswith("dict") or tp.startswith("map"):
        return "dict()"
        # return f"make({tp})"

    if tp == "str" or tp == "string":
        return '""'

    if tp == "bool":
        return "False"

    if tp == "int":
        return "0"

    if tp == "float":
        return "0.0"

    if tp == "datetime":
        return '"0001-01-01T00:00:00.000000Z"'

    return f"{tp}Factory()"


class Property:
    def __init__(self, yaml):
        self.name = yaml["name"]
        self.type = yaml["type"]

    def Default(self):
        return typeDefault(self.type)

    def CapitalizedName(self):
        return capitalize(self.name)

    def IsMap(self):
        if len(self.type) < 3:
            return False

        return self.type[:3] == "map"

    def IsArray(self):
        if len(self.type) < 2:
            return False

        return self.type[:2] == "[]"

    def SubType(self):
        if self.IsMap() or self.IsArray():
            return self.type.split("]")[1]

        return self.type

    def IsComplexType(self):
        return self.SubType() not in BASIC_TYPES

    def StrippedType(self):
        if self.IsMap():
            return "dict"
        if self.IsArray():
            return "list"
        if self.type == "string":
            return "str"
        return self.type

    def ComplexTypeValueDefault(self):
        if self.type.startswith("[]"):
            return typeDefault(self.type[2:])

        if self.type.startswith("map"):
            closing_bracket_index = self.type.find("]")
            if closing_bracket_index == -1:
                raise Exception("invalid map type: {}".format(self.type))

            return typeDefault(self.type[closing_bracket_index + 1 :])

        raise Exception("unknown type: {}".format(self.type))

    def StrippedDefault(self):
        return typeDefault(self.StrippedType())


class Struct:
    def __init__(self, yaml):
        self.name = yaml["name"]
        self.properties = [Property(p) for p in yaml["properties"]]


class Resource:
    def __init__(self, yaml):
        self.name = yaml["name"]
        
        self.primary_key = "metadata.identity"
        if "primarykey" in yaml:
            self.primary_key = yaml["primarykey"]

        self.external = None
        if "external" in yaml:
            self.external = yaml["external"]
        
        self.internal = None
        if "internal" in yaml:
            self.internal = yaml["internal"]

    def PrimaryKeyFunctionCaller(self):
        tok = self.primary_key.split(".")
        return ".".join(["{}()".format(capitalize(t)) for t in tok])
    
    def IdentityPrefix(self):
        return self.name.lower()



import json

import logging

from pystorz.internal import constants
from pystorz.store import utils
from pystorz.store import store
from markupsafe import Markup

from flask import Flask, render_template


log = logging.getLogger(__name__)


def _json_response(code: int, data):
    try:
        ret = json.dumps(data)
        log.debug("""json response: 
    {}""".format(utils.pp(data)))

        return ret, code, {"Content-Type": "application/json"}
    except Exception as e:
        log.debug("json error: {}".format(e))
        return _error_response(500, str(e))


def _error_response(code: int, message: str):
    return _json_response(code, {"error": message})


def _handle_exceptions(e):
    # traceback.print_stack()

    error_code = 500
    msg = str(e)
    if msg == constants.ErrNoSuchObject:
        error_code = 404
    if msg == constants.ErrInvalidMethod:
        error_code = 405

    return _error_response(error_code, msg)


def format_json(value):
    # return Markup(value)
    return Markup(value.replace("'", "`"))


class Server:

    def __init__(self, schema: store.SchemaHolder, thestore: store.Store):
        self.Schema = schema
        self.Store = thestore

        template_folder = utils.runtime_dir() + "/browse/templates/"
        # log.info(template_folder)

        self.app = Flask(
            "pystorz",
            template_folder=template_folder)
        # self.app.register_error_handler(Exception, _handle_exceptions)

        self._register_handlers()

    def Serve(self, host: str, port: int):
        # launch a flask server to serve the html
        log.info("serving on {}:{}".format(host, port))

        self.app.run(host, port)
        # return self.Stop

    def _register_handlers(self):
        self._register_handler(
            "/", "index", self.make_index_handler(), methods=["GET"])
        self._register_handler(
            "/id/<id>",
            "id_handler",
            self.make_id_handler(),
            methods=["GET"],
        )

        for k in self.Schema.Types():
            self._register_handler(
                f"/{k.lower()}/<pkey>",
                f"{k}_handler",
                self.make_object_handler(k),
                methods=["GET"],
            )

            type_handler = self.make_type_handler(k)
            self._register_handler(
                f"/{k.lower()}", f"{k}_type_handler1", type_handler, methods=["GET"]
            )
            self._register_handler(
                f"/{k.lower()}/", f"{k}_type_handler2", type_handler, methods=["GET"]
            )

    def make_index_handler(self):
        def handler():
            return render_template(
                "index.html",
                types=self.Schema.Types())

        return handler

    def make_id_handler(self):
        def handler(id):
            robject = self.Store.Get(store.ObjectIdentity(id))
            return render_template(
                "viewer.html",
                title=id,
                data=format_json(robject.ToJson()),
            )

        return handler

    def make_object_handler(self, t):
        def handler(pkey):
            robject = self.Store.Get(
                store.ObjectIdentity(
                    "{}/{}".format(t.lower(), pkey)))
            return render_template(
                "viewer.html",
                title=t,
                data=format_json(robject.ToJson()),
            )

        return handler

    def make_type_handler(self, t):
        def handler():
            opts = []
            ret = self.Store.List(
                store.ObjectIdentity("{}/".format(t.lower())),
                *opts
            )

            ret = [r.ToDict() for r in ret]
            return render_template(
                "viewer.html",
                title=t,
                data=format_json(json.dumps(ret)),
            )

        return handler

    def _register_handler(self, path: str, name: str, func, methods: list):
        log.info("registering {} at {} with {}".format(name, path, methods))

        self.app.add_url_rule(path, name, func, methods=methods)
